import logging
import os

import requests

from klempner import compat, errors

URL_SCHEME_MAP = {
    5672: 'amqp',  # https://www.rabbitmq.com/uri-spec.html
    21: 'ftp',  # https://tools.ietf.org/html/rfc1738
    70: 'gopher',  # https://tools.ietf.org/html/rfc4266
    80: 'http',  # https://tools.ietf.org/html/rfc7230#section-2.7.1
    443: 'https',  # https://tools.ietf.org/html/rfc7230#section-2.7.2
    1344: 'icap',  # https://tools.ietf.org/html/rfc3507#section-4.2
    631: 'ipp',  # https://tools.ietf.org/html/rfc3510
    389: 'ldap',  # https://tools.ietf.org/html/rfc4516
    636: 'ldaps',  # https://tools.ietf.org/html/rfc4516
    # https://docs.mongodb.com/manual/reference/connection-string/
    27017: 'mongodb',
    3306: 'mysql',
    119: 'nntp',  # https://tools.ietf.org/html/rfc5538
    110: 'pop',  # https://tools.ietf.org/html/rfc2384
    # https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-CONNSTRING
    5432: 'postgresql',
    6379: 'redis',  # https://www.iana.org/assignments/uri-schemes/prov/redis
    873: 'rsync',  # https://tools.ietf.org/html/rfc5781
    554: 'rtsp',  # https://tools.ietf.org/html/rfc7826#section-4.2
    322: 'rtsps',  # https://tools.ietf.org/html/rfc7826#section-4.2
    25: 'smtp',  # https://tools.ietf.org/html/draft-melnikov-smime-msa-to-mda
    161: 'snmp',  # https://tools.ietf.org/html/rfc4088
    22: 'ssh',  # https://tools.ietf.org/html/draft-ietf-secsh-scp-sftp-ssh-uri
    23: 'telnet',  # https://tools.ietf.org/html/rfc4248
    69: 'tftp',  # https://tools.ietf.org/html/rfc3617
    3372: 'tip',  # https://tools.ietf.org/html/rfc2371
    5900: 'vnc',  # https://tools.ietf.org/html/rfc7869
    602: 'xmlrpc.beep',  # https://tools.ietf.org/html/rfc3529#section-5.1
}
"""Mapping of port number to URL scheme.

This dictionary is used to identify the URL scheme based on the port number
when a port number is available.  Users of the library MAY modify the content
of this dictionary **at any time**.

"""


class DiscoveryMethod(object):
    """Available discovery methods."""

    SIMPLE = 'simple'
    """Build URLs using service name as host name."""

    CONSUL = 'consul'
    """Build consul-based service URLs."""

    CONSUL_AGENT = 'consul+agent'
    """Build consul-based service URLs using a consul agent."""

    ENV_VARS = 'environment'
    """Build URLs based on _HOST, _PORT, and _SCHEME environment variables."""

    K8S = 'kubernetes'
    """Build Kubernetes cluster-based service URLs."""

    DEFAULT = SIMPLE

    UNSET = object()

    AVAILABLE = (CONSUL, CONSUL_AGENT, ENV_VARS, K8S, SIMPLE, UNSET)


_discovery_method = DiscoveryMethod.UNSET
_discovery_parameters = {}


def reset():
    """Reset URL construction parameters."""
    configure(DiscoveryMethod.UNSET)


def ensure_configured():
    """Configure from the environment if currently unconfigured."""
    if _discovery_method is DiscoveryMethod.UNSET:
        configure_from_environment()


def configure(discovery_method, **parameters):
    """Configure the discovery method directly.

    :param discovery_method: method to use
    :param parameters: parameters required for the selected
        method
    :raises: :exc:`klempner.errors.ConfigurationError` if a required
        parameter is not provided

    """
    logger = logging.getLogger(__package__).getChild('configure')

    def require_parameter(name):
        try:
            return parameters.pop(name)
        except KeyError:
            logger.error('parameter %s is required by discovery method %s',
                         name, discovery_method)
            raise errors.ConfigurationError(name, None)

    logger.debug('configuring for discovery_method %s with parameters=%r',
                 discovery_method, parameters)
    incoming_parameters = {}
    if discovery_method == DiscoveryMethod.CONSUL:
        incoming_parameters['datacenter'] = require_parameter('datacenter')
    elif discovery_method == DiscoveryMethod.CONSUL_AGENT:
        incoming_parameters['datacenter'] = require_parameter('datacenter')
    elif discovery_method == DiscoveryMethod.K8S:
        incoming_parameters['namespace'] = require_parameter('namespace')
    elif discovery_method not in DiscoveryMethod.AVAILABLE:
        raise errors.ConfigurationError('discovery_style', discovery_method)

    global _discovery_method, _discovery_parameters
    if _discovery_method is DiscoveryMethod.UNSET:
        logger.info('setting discovery method to %r with parameters=%r',
                    discovery_method, incoming_parameters)
    else:
        logger.info(
            'setting discovery method: current_method=%r new_method=%r '
            'parameters=%r', _discovery_method, discovery_method,
            incoming_parameters)
    _discovery_method = discovery_method
    _discovery_parameters.clear()
    _discovery_parameters.update(incoming_parameters)
    if parameters:
        logger.warning(
            'discovery style %s does not accept additional parameters, '
            '%d extra parameters were passed to configure', discovery_method,
            len(parameters))


def configure_from_environment():
    """Set the discovery method from ``$KLEMPNER_DISCOVERY``.

    This method configures the library based on the
    :envvar:`KLEMPNER_DISCOVERY` environment variable.  If the
    environment variable is not set, then the "simple" configuration
    is used.

    """
    logger = logging.getLogger(__package__).getChild(
        'configure_from_environment')
    new_method = os.environ.get('KLEMPNER_DISCOVERY', DiscoveryMethod.SIMPLE)
    logger.debug('configuring from environment: discovery_method=%s',
                 new_method)

    def require_envvar(name):
        try:
            return os.environ[name]
        except KeyError:
            logger.error(
                'discovery method %s requires the %s environment variable',
                new_method, name)
            raise errors.ConfigurationError(name, None)

    parameters = {}
    if new_method == DiscoveryMethod.CONSUL:
        parameters['datacenter'] = require_envvar('CONSUL_DATACENTER')
    elif new_method == DiscoveryMethod.CONSUL_AGENT:
        url = compat.urlunparse(('http', require_envvar('CONSUL_HTTP_ADDR'),
                                 '/v1/agent/self', None, None, None))
        response = requests.get(url)
        response.raise_for_status()
        body = response.json()
        parameters['datacenter'] = body['Config']['Datacenter']
    elif new_method == DiscoveryMethod.K8S:
        parameters['namespace'] = os.environ.get('KUBERNETES_NAMESPACE',
                                                 'default')
    elif new_method not in DiscoveryMethod.AVAILABLE:
        raise errors.ConfigurationError('discovery_style', new_method)

    configure(new_method, **parameters)


def get_discovery_details():
    """Retrieve the configured method and parameters.

    :rtype: tuple(str, dict)

    """
    return _discovery_method, _discovery_parameters.copy()



class KlempnerError(Exception):
    """Library root exception class."""
    pass


class ServiceNotFoundError(KlempnerError):
    """Requested service was not found."""

    def __init__(self, service_name, *args, **kwargs):
        self.service_name = service_name
        super(ServiceNotFoundError, self).__init__(service_name, *args,
                                                   **kwargs)


class ConfigurationError(KlempnerError):
    """Configuration is invalid."""

    def __init__(self, config_option, config_value, *args):
        self.configuration_name = config_option
        self.configuration_value = config_value
        super(ConfigurationError, self).__init__(config_option, config_value,
                                                 *args)



import logging
import unittest

from klempner import config, errors, url

import tests.helpers


class DirectConfigTests(unittest.TestCase):
    def tearDown(self):
        super(DirectConfigTests, self).tearDown()
        config.reset()

    def test_that_unknown_discovery_method_raises_configuration_error(self):
        with self.assertRaises(errors.ConfigurationError) as context:
            config.configure('unknown-option')
        self.assertEqual('discovery_style',
                         context.exception.configuration_name)
        self.assertEqual('unknown-option',
                         context.exception.configuration_value)

    def test_that_simple_discovery_is_parameterless(self):
        config.configure(config.DiscoveryMethod.SIMPLE)
        _, parameters = config.get_discovery_details()
        self.assertEqual({}, parameters)

    def test_that_consul_discovery_without_datacenter_fails(self):
        with self.assertRaises(errors.ConfigurationError) as context:
            config.configure(config.DiscoveryMethod.CONSUL)
        self.assertEqual('datacenter', context.exception.configuration_name)
        self.assertIs(None, context.exception.configuration_value)

    def test_that_consul_agent_discovery_without_datacenter_fails(self):
        with self.assertRaises(errors.ConfigurationError) as context:
            config.configure(config.DiscoveryMethod.CONSUL_AGENT)
        self.assertEqual('datacenter', context.exception.configuration_name)
        self.assertIs(None, context.exception.configuration_value)


class ConfigByEnvironTests(tests.helpers.EnvironmentMixin, unittest.TestCase):
    def tearDown(self):
        super(ConfigByEnvironTests, self).tearDown()
        config.reset()

    def test_that_unknown_discovery_method_raises_configuration_error(self):
        self.setenv('KLEMPNER_DISCOVERY', 'unknown-option')
        with self.assertRaises(errors.ConfigurationError) as context:
            config.configure_from_environment()
        self.assertEqual('discovery_style',
                         context.exception.configuration_name)
        self.assertEqual('unknown-option',
                         context.exception.configuration_value)

    def test_that_consul_discovery_without_datacenter_fails(self):
        self.setenv('KLEMPNER_DISCOVERY', config.DiscoveryMethod.CONSUL)
        self.unsetenv('CONSUL_DATACENTER')
        with self.assertRaises(errors.ConfigurationError) as context:
            config.configure_from_environment()
        self.assertEqual('CONSUL_DATACENTER',
                         context.exception.configuration_name)
        self.assertIs(None, context.exception.configuration_value)

    def test_that_consul_agent_discovery_without_http_addr_fails(self):
        self.setenv('KLEMPNER_DISCOVERY', config.DiscoveryMethod.CONSUL_AGENT)
        self.unsetenv('CONSUL_HTTP_ADDR')
        with self.assertRaises(errors.ConfigurationError) as context:
            config.configure_from_environment()
        self.assertEqual('CONSUL_HTTP_ADDR',
                         context.exception.configuration_name)
        self.assertIs(None, context.exception.configuration_value)


class ErrorHandlingTests(unittest.TestCase):
    def test_that_warning_is_logged_for_unused_parameters(self):
        log_messages = []

        class Recorder(logging.Handler):
            def emit(self, record):
                log_messages.append(record.getMessage())

        logger = logging.getLogger('klempner.configure')
        handler = Recorder()
        logger.addHandler(handler)
        self.addCleanup(logger.removeHandler, handler)

        config.configure(config.DiscoveryMethod.SIMPLE, unusued='parameter')
        for message in log_messages:
            if 'simple does not accept additional parameters' in message:
                break
        else:
            self.fail('expected to find warning message in ' +
                      repr(log_messages))


class SchemeMappingTests(tests.helpers.EnvironmentMixin, unittest.TestCase):
    def setUp(self):
        self._saved_scheme_map = config.URL_SCHEME_MAP.copy()
        super(SchemeMappingTests, self).setUp()

    def tearDown(self):
        super(SchemeMappingTests, self).tearDown()
        config.URL_SCHEME_MAP.clear()
        config.URL_SCHEME_MAP.update(self._saved_scheme_map)

    def test_that_mapping_can_be_disabled(self):
        config.URL_SCHEME_MAP.clear()
        config.configure(config.DiscoveryMethod.ENV_VARS)
        self.setenv('ACCOUNT_HOST', 'account.example.com')
        self.setenv('ACCOUNT_PORT', '443')
        self.assertEqual('http://account.example.com:443/',
                         url.build_url('account'))

    def test_that_mapping_can_be_overridden(self):
        config.URL_SCHEME_MAP[5672] = 'rabbitmq'
        config.configure(config.DiscoveryMethod.ENV_VARS)
        self.setenv('ACCOUNT_HOST', 'account.example.com')
        self.setenv('ACCOUNT_PORT', '5672')
        self.assertEqual('rabbitmq://account.example.com:5672/',
                         url.build_url('account'))



from __future__ import unicode_literals

import logging
import os
try:
    from urllib.parse import urlsplit
except ImportError:
    from urlparse import urlsplit

import requests.adapters

import cachetools
from klempner import compat, config, errors

#    pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
#    sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
#                  / "*" / "+" / "," / ";" / "="
#    unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"

PATH_SAFE_CHARS = ":@!$&'()*+,;=-._~"
"""Safe characters for path elements."""


class ConsulAgentAdapter(requests.adapters.HTTPAdapter):
    """Adapter that sends `consul://` requests to a consul agent."""

    def send(self, request, stream=False, timeout=None, verify=True, cert=None,
             proxies=None):
        _, host, path, params, query, fragment = compat.urlparse(request.url)
        path = host + path
        request.url = compat.urlunparse(
            ('http', os.environ['CONSUL_HTTP_ADDR'], path, params, query,
             fragment))
        return super(ConsulAgentAdapter, self).send(
            request, stream=stream, timeout=timeout, verify=verify, cert=cert,
            proxies=proxies)  # yapf: disable


class State(object):
    """Module state.

    A single instance of this class exists as a module-level property.
    It caches information as it is discovered.  Applications SHOULD
    call :func:`.reset_cache` if they suspect that the discovery
    configuration has changed.

    """

    def __init__(self):
        self.discovery_cache = cachetools.TTLCache(50, 300)
        self.logger = logging.getLogger(__package__)
        self.session = requests.Session()
        self.session.mount('consul://', ConsulAgentAdapter())

    def clear(self):
        self.discovery_cache.clear()
        self.session.close()


_state = State()


def reset_cache():
    """Reset internal caches.

    Applications MUST call this function if they have changed discovery
    configuration details or suspect that they may have changed.  This
    should not happen often since the discovery configuration is based
    primarily on environment variables which are not modifiable from
    outside of the process.

    """
    config.reset()
    _state.clear()


def build_url(service, *path, **query):
    """Build a URL that targets `service`.

    :param str service: service to target
    :param path: request path elements
    :param query: request query parameters
    :returns: a fully-formed, absolute URL
    :rtype: str

    """
    config.ensure_configured()
    buf = compat.StringIO()
    _write_network_portion(buf, service)
    buf.write('/')
    buf.write('/'.join(
        compat.quote(str(p), safe=PATH_SAFE_CHARS) for p in path))

    query_tuples = []
    for name, value in query.items():
        if isinstance(value, compat.Mapping):
            raise ValueError('Mapping query parameters are unsupported')
        if (isinstance(value, compat.Iterable)
                and not isinstance(value, compat.TEXT_TYPES)):
            query_tuples.extend((name, elm) for elm in sorted(value))
        else:
            query_tuples.append((name, value))
    if query_tuples:
        query_tuples.sort()
        buf.write('?')
        buf.write('&'.join(
            '{0}={1}'.format(_quote_query_arg(name), _quote_query_arg(value))
            for name, value in query_tuples))

    return buf.getvalue()


def _write_network_portion(buf, service):
    """Add the discovered network portion to `buf`.

    :param klempner.compat.StringIO buf: buffer to write the discovered
        network details to
    :param str service: name of the service that is being looked up

    """
    env_service = service.upper()
    discovery_style, parameters = config.get_discovery_details()
    if discovery_style == config.DiscoveryMethod.CONSUL:
        buf.write('http://')
        buf.write(service)
        buf.write('.service.')
        buf.write(parameters['datacenter'])
        buf.write('.consul')
    elif discovery_style == config.DiscoveryMethod.CONSUL_AGENT:
        sentinel = object()
        body = _state.discovery_cache.get(service, sentinel)
        if body is sentinel:
            response = _state.session.get(
                'consul://v1/catalog/service/{0}'.format(service))
            response.raise_for_status()
            body = response.json()
            _state.discovery_cache[service] = body

        if not body:  # service does not exist in consul
            raise errors.ServiceNotFoundError(service)
        else:
            calculated_scheme = config.URL_SCHEME_MAP.get(
                body[0]['ServicePort'], 'http')
            meta = body[0].get('ServiceMeta', {})
            buf.write(meta.get('protocol', calculated_scheme))
            buf.write('://')
            buf.write(
                '{ServiceName}.service.{Datacenter}.consul'.format(**body[0]))
            buf.write(':' + str(body[0]['ServicePort']))
    elif discovery_style == config.DiscoveryMethod.K8S:
        buf.write('http://')
        buf.write(service + '.')
        buf.write(parameters['namespace'])
        buf.write('.svc.cluster.local')
    elif discovery_style == config.DiscoveryMethod.ENV_VARS:
        scheme = os.environ.get('{0}_SCHEME'.format(env_service), None)
        host = os.environ.get('{0}_HOST'.format(env_service), None)
        port = os.environ.get('{0}_PORT'.format(env_service), None)

        if port is not None and port.startswith('tcp://'):
            # special case for docker's ip:port format
            parts = urlsplit(port)
            port = str(parts.port)
            if host is None:
                host = parts.hostname
        if scheme is None:
            if port is not None:
                scheme = config.URL_SCHEME_MAP.get(int(port), 'http')
            else:
                scheme = 'http'
        buf.write(scheme)
        buf.write('://')
        buf.write(host or service)
        if port is not None:
            buf.write(':')
            buf.write(port)
    else:
        buf.write('http://')
        buf.write(service)


def _quote_query_arg(v):
    if not isinstance(v, compat.TEXT_TYPES):
        v = str(v)
    return compat.quote(v.encode('utf-8'))



"""Python 2/3 compatibility shim."""
try:
    from io import StringIO
    from urllib.parse import quote, urlparse, urlunparse
except ImportError:  # pragma: no cover
    from StringIO import StringIO
    from urllib import quote
    from urlparse import urlparse, urlunparse

try:
    from typing import Iterable, Mapping
except ImportError:  # pragma: no cover
    from collections import Iterable, Mapping

try:
    TEXT_TYPES = (str, unicode)
except NameError:  # pragma: no cover
    TEXT_TYPES = (str, )

__all__ = [
    'Iterable',
    'Mapping',
    'quote',
    'StringIO',
    'TEXT_TYPES',
    'urlparse',
    'urlunparse',
]



from __future__ import unicode_literals

import unittest

import klempner.url


class QueryParameterTests(unittest.TestCase):
    def setUp(self):
        super(QueryParameterTests, self).setUp()
        klempner.url.reset_cache()

    def test_that_query_parameters_are_encoded(self):
        url = klempner.url.build_url('some-service', arg1='value',
                                     arg2='with space',
                                     arg3='r\u00E9sum\u00E9')
        self.assertEqual(
            'http://some-service/'
            '?arg1=value&arg2=with%20space&arg3=r%C3%A9sum%C3%A9',
            url,
        )

    def test_that_sequences_are_expanded(self):
        url = klempner.url.build_url(
            'some-service',
            list_arg=['one', 'two'],
            set_arg=set(['one', 'two']),
            tuple_arg=('one', 'two'),
        )
        self.assertEqual(
            'http://some-service/'
            '?list_arg=one&list_arg=two'
            '&set_arg=one&set_arg=two'
            '&tuple_arg=one&tuple_arg=two',
            url,
        )

    def test_that_mapping_query_params_fail(self):
        with self.assertRaises(ValueError):
            klempner.url.build_url('some-service', val={'one': 1, 'two': '2'})


class PathTests(unittest.TestCase):
    def setUp(self):
        super(PathTests, self).setUp()
        klempner.url.reset_cache()

    def test_that_path_elements_are_quoted(self):
        url = klempner.url.build_url(
            'some-service', 'with spaces', 'other:interesting@chars',
            '(sub!&*+,;=!delims)', 'unreserved-._~chars', 'quoted<>{}/chars')
        self.assertEqual(
            'http://some-service'
            '/with%20spaces'
            '/other:interesting@chars'
            '/(sub!&*+,;=!delims)'
            '/unreserved-._~chars'
            '/quoted%3C%3E%7B%7D%2Fchars',
            url,
        )

    def test_that_non_string_path_elements_are_stringified(self):
        url = klempner.url.build_url('some-service', 1234, True, None)
        self.assertEqual('http://some-service/1234/True/None', url)

    def test_that_no_path_always_ends_with_slash(self):
        self.assertEqual('http://some-service/',
                         klempner.url.build_url('some-service'))
        self.assertEqual('http://some-service/?q=1',
                         klempner.url.build_url('some-service', q=1))



from __future__ import unicode_literals

import unittest

from tests import helpers
import klempner.config
import klempner.url


class SimpleEnvironmentTests(helpers.EnvironmentMixin, unittest.TestCase):
    def setUp(self):
        super(SimpleEnvironmentTests, self).setUp()
        klempner.url.reset_cache()
        self.setenv('KLEMPNER_DISCOVERY',
                    klempner.config.DiscoveryMethod.ENV_VARS)

    def test_that_scheme_envvar_is_honored(self):
        self.setenv('ACCOUNT_SCHEME', 'https')
        url = klempner.url.build_url('account')
        self.assertEqual('https://account/', url)

    def test_that_host_envvar_is_honored(self):
        self.setenv('ACCOUNT_HOST', '10.2.12.23')
        url = klempner.url.build_url('account')
        self.assertEqual('http://10.2.12.23/', url)

    def test_that_port_envvar_is_honored(self):
        self.setenv('ACCOUNT_PORT', '11223')
        url = klempner.url.build_url('account')
        self.assertEqual('http://account:11223/', url)

    def test_that_docker_linked_port_is_accepted(self):
        self.setenv('ACCOUNT_PORT', 'tcp://10.2.12.23:11223')
        url = klempner.url.build_url('account')
        self.assertEqual('http://10.2.12.23:11223/', url)

    def test_that_host_is_taken_from_docker_linked_port(self):
        self.setenv('ACCOUNT_HOST', '10.2.12.23')
        self.setenv('ACCOUNT_PORT', 'tcp://10.2.12.24:11223')
        url = klempner.url.build_url('account')
        self.assertEqual('http://10.2.12.23:11223/', url)

    def test_that_scheme_is_set_from_port(self):
        self.setenv('ACCOUNT_HOST', '10.2.12.23')
        self.setenv('ACCOUNT_PORT', '443')
        url = klempner.url.build_url('account')
        self.assertEqual('https://10.2.12.23:443/', url)

    def test_that_scheme_is_taken_from_docker_linked_port(self):
        self.setenv('ACCOUNT_PORT', 'tcp://10.2.12.23:443')
        url = klempner.url.build_url('account')
        self.assertEqual('https://10.2.12.23:443/', url)



from __future__ import unicode_literals

import unittest
import uuid

import klempner.config
from tests import helpers
import klempner.url


class SimpleK8sTests(helpers.EnvironmentMixin, unittest.TestCase):
    def setUp(self):
        super(SimpleK8sTests, self).setUp()
        klempner.url.reset_cache()
        self.setenv('KLEMPNER_DISCOVERY', klempner.config.DiscoveryMethod.K8S)
        self.setenv('KUBERNETES_NAMESPACE', 'development')

    def test_that_k8s_namespace_environment_sets_datacenter_name(self):
        env = str(uuid.uuid4())
        self.setenv('KUBERNETES_NAMESPACE', env)
        self.assertEqual('http://account.{0}.svc.cluster.local/'.format(env),
                         klempner.url.build_url('account'))

    def test_that_k8s_namespace_defaults_when_envvar_is_not_set(self):
        self.unsetenv('KUBERNETES_NAMESPACE')
        self.assertEqual('http://account.default.svc.cluster.local/',
                         klempner.url.build_url('account'))



#!/usr/bin/env python

import setuptools

import klempner

setuptools.setup(
    name='klempner',
    version=klempner.version,
    description='Construct service request URLs',
    long_description=open('README.rst').read(),
    url='https://github.com/dave-shawley/klempner',
    author='Dave Shawley',
    author_email='daveshawley@gmail.com',
    packages=['klempner'],
    install_requires=[
        'cachetools==3.1.0',
        'requests==2.21.0',
    ],
    extras_require={
        'dev': [
            'coverage==4.5.3',
            'flake8==3.7.7',
            'tox==3.8.6',
            'yapf==0.26.0',
        ],
        'docs': [
            'sphinx==2.0.1',
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
    ],
)



import os
import unittest


class EnvironmentMixin(unittest.TestCase):
    def setUp(self):
        super(EnvironmentMixin, self).setUp()
        self._environment = {}

    def tearDown(self):
        super(EnvironmentMixin, self).tearDown()
        for name, value in self._environment.items():
            if value is None:
                os.environ.pop(name, None)
            else:
                os.environ[name] = value

    def setenv(self, name, value):
        self._environment.setdefault(name, os.environ.get(name, None))
        os.environ[name] = value

    def unsetenv(self, name):
        self._environment.setdefault(name, os.environ.get(name, None))
        os.environ.pop(name, None)



from __future__ import unicode_literals

import os
import random
import unittest
import uuid

import requests

import klempner.config
from tests import helpers
import klempner.errors
import klempner.url


class SimpleConsulTests(helpers.EnvironmentMixin, unittest.TestCase):
    def test_that_consul_datacenter_environment_sets_datacenter_name(self):
        klempner.url.reset_cache()
        self.setenv('KLEMPNER_DISCOVERY',
                    klempner.config.DiscoveryMethod.CONSUL)
        env = str(uuid.uuid4())
        self.setenv('CONSUL_DATACENTER', env)
        self.assertEqual('http://account.service.{0}.consul/'.format(env),
                         klempner.url.build_url('account'))


class AgentBasedTests(helpers.EnvironmentMixin, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(AgentBasedTests, cls).setUpClass()
        cls.session = requests.Session()

        try:
            cls.agent_url = 'http://{0}/v1/agent'.format(
                os.environ['CONSUL_HTTP_ADDR'])
        except KeyError:
            raise unittest.SkipTest('Consul agent is not present')

        response = cls.session.get(cls.agent_url + '/self')
        response.raise_for_status()
        body = response.json()
        cls.datacenter = body['Config']['Datacenter']

    def setUp(self):
        super(AgentBasedTests, self).setUp()
        klempner.url.reset_cache()
        self.setenv('KLEMPNER_DISCOVERY',
                    klempner.config.DiscoveryMethod.CONSUL_AGENT)
        self.unsetenv('CONSUL_DATACENTER')
        self._service_ids = set()

    def tearDown(self):
        super(AgentBasedTests, self).tearDown()
        for service_id in list(self._service_ids):
            self.deregister_service(service_id, ignore_error=True)

    def register_service(self, meta=None, port=None):
        service_id = str(uuid.uuid4())
        service_details = {
            'ID': service_id,
            'Name': 's' + service_id.replace('-', '').lower(),
            'Address': '10.0.0.1',
            'Port': port or random.randint(10000, 20000),
            'Datacenter': self.datacenter,
        }
        if meta:
            service_details['Meta'] = meta
        response = self.session.put(self.agent_url + '/service/register',
                                    json=service_details)
        response.raise_for_status()
        self._service_ids.add(service_id)
        return service_details

    def deregister_service(self, service_id, ignore_error=False):
        response = self.session.put(
            self.agent_url + '/service/deregister/{0}'.format(service_id))
        if not ignore_error:
            response.raise_for_status()
        self._service_ids.discard(service_id)

    def test_that_details_are_fetched_from_consul_agent(self):
        service_info = self.register_service()
        self.assertEqual(
            'http://{Name}.service.{Datacenter}.consul:{Port}/'.format(
                **service_info),
            klempner.url.build_url(service_info['Name']),
        )

    def test_that_nonexistent_service_raises_exception(self):
        with self.assertRaises(klempner.errors.ServiceNotFoundError):
            klempner.url.build_url(str(uuid.uuid4()))

    def test_that_service_lookup_is_cached(self):
        service_info = self.register_service()
        expected = 'http://{Name}.service.{Datacenter}.consul:{Port}/'.format(
            **service_info)
        self.assertEqual(expected,
                         klempner.url.build_url(service_info['Name']))

        self.deregister_service(service_info['ID'])
        self.assertEqual(expected,
                         klempner.url.build_url(service_info['Name']))

    def test_that_scheme_set_by_metadata(self):
        service_info = self.register_service(meta={'protocol': 'https'})
        expected = 'https://{Name}.service.{Datacenter}.consul:{Port}/'.format(
            **service_info)
        self.assertEqual(expected,
                         klempner.url.build_url(service_info['Name']))

    def test_that_scheme_set_by_port(self):
        for port, scheme in klempner.config.URL_SCHEME_MAP.items():
            service_info = self.register_service(port=port)
            self.assertEqual(
                '{scheme}://{Name}.service.{Datacenter}.consul:{port}/'.format(
                    port=port, scheme=scheme, **service_info),
                klempner.url.build_url(service_info['Name']))
